@@ -1,5 +1,5 @@
 /* kwsearch.c - searching subroutines using kwset for grep.
-   Copyright 1992, 1998, 2000, 2007, 2009-2014 Free Software Foundation, Inc.
+   Copyright 1992, 1998, 2000, 2007, 2009-2015 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -37,14 +37,10 @@ void
 Fcompile (char const *pattern, size_t size)
 {
   size_t total = size;
-  mb_len_map_t *map = NULL;
-  char const *pat = (match_icase && MB_CUR_MAX > 1
-                     ? mbtoupper (pattern, &total, &map)
-                     : pattern);
 
   kwsinit (&kwset);
 
-  char const *p = pat;
+  char const *p = pattern;
   do
     {
       size_t len;
@@ -81,26 +77,6 @@ Fcompile (char const *pattern, size_t size)
   kwsprep (kwset);
 }
 
-/* Apply the MAP (created by mbtoupper) to the uppercase-buffer-relative
-   *OFF and *LEN, converting them to be relative to the original buffer.  */
-
-static void
-mb_case_map_apply (mb_len_map_t const *map, size_t *off, size_t *len)
-{
-  if (map)
-    {
-      size_t off_incr = 0;
-      size_t len_incr = 0;
-      size_t k;
-      for (k = 0; k < *off; k++)
-        off_incr += map[k];
-      for (; k < *off + *len; k++)
-        len_incr += map[k];
-      *off += off_incr;
-      *len += len_incr;
-    }
-}
-
 size_t
 Fexecute (char const *buf, size_t size, size_t *match_size,
           char const *start_ptr)
@@ -110,18 +86,6 @@ Fexecute (char const *buf, size_t size, size_t *match_size,
   char eol = eolbyte;
   struct kwsmatch kwsmatch;
   size_t ret_val;
-  mb_len_map_t *map = NULL;
-
-  if (MB_CUR_MAX > 1)
-    {
-      if (match_icase)
-        {
-          char *case_buf = mbtoupper (buf, &size, &map);
-          if (start_ptr)
-            start_ptr = case_buf + (start_ptr - buf);
-          buf = case_buf;
-        }
-    }
 
   for (mb_start = beg = start_ptr ? start_ptr : buf; beg <= buf + size; beg++)
     {
@@ -129,24 +93,40 @@ Fexecute (char const *buf, size_t size, size_t *match_size,
                                buf + size - beg + match_lines, &kwsmatch);
       if (offset == (size_t) -1)
         goto failure;
-      len = kwsmatch.size[0] - match_lines;
+      len = kwsmatch.size[0] - 2 * match_lines;
       if (!match_lines && MB_CUR_MAX > 1 && !using_utf8 ()
           && mb_goback (&mb_start, beg + offset, buf + size) != 0)
         {
-          /* The match was a part of multibyte character, advance at least
-             one byte to ensure no infinite loop happens.  */
-          beg = mb_start;
+          /* We have matched a single byte that is not at the beginning of a
+             multibyte character.  mb_goback has advanced MB_START past that
+             multibyte character.  Now, we want to position BEG so that the
+             next kwsexec search starts there.  Thus, to compensate for the
+             for-loop's BEG++, above, subtract one here.  This code is
+             unusually hard to reach, and exceptionally, let's show how to
+             trigger it here:
+
+               printf '\203AA\n'|LC_ALL=ja_JP.SHIFT_JIS src/grep -F A
+
+             That assumes the named locale is installed.
+             Note that your system's shift-JIS locale may have a different
+             name, possibly including "sjis".  */
+          beg = mb_start - 1;
           continue;
         }
       beg += offset;
       if (start_ptr && !match_words)
         goto success_in_beg_and_len;
       if (match_lines)
-        goto success_in_beg_and_len;
+        {
+          len += start_ptr == NULL;
+          goto success_in_beg_and_len;
+        }
       if (match_words)
         for (try = beg; ; )
           {
-            if (wordchar (mb_prev_wc (buf, try, buf + size)))
+            char const *bol = memrchr (buf, eol, beg - buf);
+            bol = bol ? bol + 1 : buf;
+            if (wordchar (mb_prev_wc (bol, try, buf + size)))
               break;
             if (wordchar (mb_next_wc (try + len, buf + size)))
               {
@@ -171,16 +151,13 @@ Fexecute (char const *buf, size_t size, size_t *match_size,
   return -1;
 
  success:
-  if ((end = memchr (beg + len, eol, (buf + size) - (beg + len))) != NULL)
-    end++;
-  else
-    end = buf + size;
-  while (buf < beg && beg[-1] != eol)
-    --beg;
+  end = memchr (beg + len, eol, (buf + size) - (beg + len));
+  end = end ? end + 1 : buf + size;
+  beg = memrchr (buf, eol, beg - buf);
+  beg = beg ? beg + 1 : buf;
   len = end - beg;
  success_in_beg_and_len:;
   size_t off = beg - buf;
-  mb_case_map_apply (map, &off, &len);
 
   *match_size = len;
   ret_val = off;
