@@ -1,5 +1,5 @@
 /* kwset.c - search for any of a set of keywords.
-   Copyright (C) 1989, 1998, 2000, 2005, 2007, 2009-2014 Free Software
+   Copyright (C) 1989, 1998, 2000, 2005, 2007, 2009-2015 Free Software
    Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
@@ -637,18 +637,19 @@ bmexec_trans (kwset_t kwset, char const *text, size_t size)
                        delta1 when the latter doesn't advance much.  */
                     int advance_heuristic = 16 * sizeof (long);
                     if (advance_heuristic <= tp - tp0)
-                      goto big_advance;
+                      continue;
                     tp--;
                     tp = memchr_kwset (tp, text + size - tp, kwset);
                     if (! tp)
                       return -1;
                     tp++;
+                    if (ep <= tp)
+                      break;
                   }
               }
           }
         if (bm_delta2_search (&tp, ep, sp, len, trans, gc1, gc2, d1, kwset))
           return tp - text;
-      big_advance:;
       }
 
   /* Now we have only a few characters left to search.  We
